@@ -1,7 +1,9 @@
 import time
 from collections import OrderedDict
 
+from django.contrib.admin.templatetags.admin_urls import admin_urlname
 from django.http import HttpResponseRedirect
+from django.shortcuts import render
 from django.urls import reverse
 from functools import update_wrapper
 
@@ -9,26 +11,68 @@ from django.conf import settings
 from django.contrib.admin import AdminSite
 from django.core.cache import caches
 from django.template.response import TemplateResponse
-from django.views.decorators.cache import never_cache
+from django.utils.decorators import method_decorator
+from django.views.decorators.cache import never_cache, cache_page
+from django.views.decorators.vary import vary_on_cookie
 
 from . import get_full_version
 from . import settings as smart_settings
-from .templates.templatetags.smart import as_bool
+# from .settings import process_bookmarks
+from .templatetags.smart import as_bool
+from .utils import SmartList
 
 cache = caches['default']
 
 
+def _parse_section():
+    base = smart_settings.SECTIONS
+    ret = {'_hidden_': []}
+    for name, entries in base.items():
+        ret[name] = SmartList(entries)
+    if 'Other' not in ret:
+        ret['Other'] = []
+    return ret
+
+
+@method_decorator([vary_on_cookie, cache_page(smart_settings.SYSINFO_TTL)], name='admin_sysinfo')
 class SmartAdminSite(AdminSite):
     sysinfo_url = False
+    index_template = 'admin/index.html'
+
+    def get_bookmarks(self, request):
+        if smart_settings.BOOKMARKS_PERMISSION is None or request.user.has_permission(
+                smart_settings.BOOKMARKS_PERMISSION):
+            bookmarks = []
+            values = smart_settings.get_bookmarks(request)
+
+            for entry in values:
+                if not entry:
+                    continue
+                if isinstance(entry, str):
+                    label, url, cls = entry, entry, 'viewlink'
+                elif len(entry) == 2:
+                    label, url, cls = entry[0], entry[1], 'viewlink'
+                elif len(entry) == 3:
+                    label, url, cls = entry
+                else:
+                    raise ValueError(f"Invalid entry '{entry}' for BOOKMARKS")
+                bookmarks.append([label, url, cls])
+            return bookmarks
+        return []
 
     def each_context(self, request):
         context = super().each_context(request)
         context['sysinfo'] = self.sysinfo_url
+        context['smart_settings'] = smart_settings
         context['enable_switch'] = smart_settings.ENABLE_SWITCH
-        context['quick_links'] = smart_settings.BOOKMARKS
+        context['bookmarks'] = self.get_bookmarks(request)
         context['smart'] = self.is_smart_enabled(request)
         context['smart_sections'], context['model_to_section'] = self._get_menu(request)
         context['adminsite'] = self
+        if smart_settings.PROFILE_LINK and request.user.is_authenticated:
+            context['profile_link'] = reverse(admin_urlname(request.user._meta, 'change'),
+                                              args=[request.user.pk])
+
         return context
 
     def is_smart_enabled(self, request):
@@ -39,7 +83,20 @@ class SmartAdminSite(AdminSite):
         if self.is_smart_enabled(request):
             return self.smart_index(request)
         else:
-            return super(SmartAdminSite, self).index(request)
+            return super().index(request)
+
+    def admin_sysinfo(self, request):
+        from django_sysinfo.api import get_sysinfo
+        infos = get_sysinfo(request)
+        infos.setdefault('extra', {})
+        infos.setdefault('checks', {})
+        context = self.each_context(request)
+        context.update({'title': 'sysinfo',
+                        'infos': infos,
+                        'enable_switch': True,
+                        'has_permission': True,
+                        })
+        return render(request, 'admin/sysinfo/sysinfo.html', context)
 
     def app_index(self, request, app_label, extra_context=None):
         groups, __ = self._get_menu(request)
@@ -70,15 +127,15 @@ class SmartAdminSite(AdminSite):
         urlpatterns = [path('~groups/<str:group>/', wrap(self.smart_section), name='group_list'),
                        path('smart/<str:on_off>/', wrap(self.smart_toggle), name='smart_toggle'),
                        ]
-        urlpatterns += super(SmartAdminSite, self).get_urls()
+
         try:
-            from django_sysinfo.views import admin_sysinfo
             if 'django_sysinfo' in settings.INSTALLED_APPS:
                 from django.urls import reverse_lazy
-                urlpatterns += [path('sysinfo/', wrap(admin_sysinfo), name='smart-sysinfo-admin'), ]
+                urlpatterns += [path('~sysinfo/', wrap(self.admin_sysinfo), name='smart-sysinfo-admin'), ]
                 self.sysinfo_url = reverse_lazy('admin:smart-sysinfo-admin')
         except ImportError:
             pass
+        urlpatterns += super().get_urls()
 
         return urlpatterns
 
@@ -87,17 +144,14 @@ class SmartAdminSite(AdminSite):
             ver = time.time
         else:
             ver = get_full_version
-
-        sections = smart_settings.SECTIONS
-
-        key1 = f'{hash(str(sections))}:groups:{ver()}'
-        key2 = f'{hash(str(sections))}:models:{ver()}'
+        sections = _parse_section()
+        key1 = f'{hash(repr(sections))}:groups:{ver()}'
+        key2 = f'{hash(repr(sections))}:models:{ver()}'
 
         groups = cache.get(key1)
         model_to_section = cache.get(key2)
         if not (groups and model_to_section):
             model_to_section = {}
-            sections = smart_settings.SECTIONS
             groups = OrderedDict([(k, []) for k in sections.keys()])
             app_list = self.get_app_list(request)
 
@@ -111,7 +165,7 @@ class SmartAdminSite(AdminSite):
                     if fqn in models:
                         return sec
                     elif app['app_label'] in models:
-                        target = sec
+                        return sec
                 return target
 
             for app in app_list:
